#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <netdb.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include <sys/time.h>
#include "csyncd.h"

extern int errno;
extern struct sockaddr_in sin_local, sin_remote;
extern int s;

char *dtstrings[] = {"INVALID","ARP","TCP","UDP","RAW","SOCKSTAT",
    "IPAUTOFW","IPFWD","ROUTE"};
char *opstrings[] = {"INVALID","ADD","REMOVE","FLUSH","ACK","NAK",
    "BOOTSTRAP","RECONFIG"};

/*
 * Send a packet to the passive controller, allowing 3 retries if no response
 * or NAK
 */
int send_packet(syncpkt_t *pkt)
{
    int len = sizeof(syncpkt_t);
    int retry=0,ret;
    char buf[256];

    do {
        sendto(s,pkt,len,0,(struct sockaddr *)&sin_remote,sizeof(sin_remote));
        ret=get_reply();
        if(ret < 0) {
            sprintf(buf,"No response to %s %s request. Retrying",
             dtstrings[(int)pkt->datatype], opstrings[(int)pkt->oper]);
            logstr(LOG_WARNING,buf);
            retry++;
        } else if(ret == OP_NAK) {
            sprintf(buf,"Received NAK on %s %s request. Retrying",
             dtstrings[(int)pkt->datatype], opstrings[(int)pkt->oper]);
            logstr(LOG_WARNING,buf);
            retry++;        
        } else 
            return ret;;
    } while(retry < 3);
    logstr(LOG_ERR,"All retries failed. Giving up on this packet.");
    return -1;    
}

/*
 * Get an ACK/NAK from the passive controller. If we don't get one in the
 * allowed time, log an error and give up. Returns pkt OP code on success and
 * -1 on timeout.
 */
int get_reply(void)
{
    syncpkt_t pkt;
    int len = sizeof(pkt);
    int salen = sizeof(sin_local);
    struct timeval tv;
    fd_set rfds;
    int ret;
    char buf[128];
    
    tv.tv_sec = 2; /* Wait up to 2 secs for reply */
    tv.tv_usec = 0;
    FD_ZERO(&rfds);
    FD_SET(s,&rfds);
    if((ret=select(s+1,&rfds,0,0,&tv)) > 0) {
        recvfrom(s,&pkt,len,0,(struct sockaddr *)&sin_local,&salen);
    } else if(ret < 0) {
        sprintf(buf,"select: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
        return -1;
    } else
    	return -1;
    return pkt.oper;
}

/*
 * Send an ARP ADD request to the passive controller
 */
int arp_add(arp_data_t *arp)
{
    syncpkt_t pkt;

    pkt.datatype = D_ARP;
    pkt.oper = OP_ADD;
    memcpy(&pkt.msg,(char *)arp,sizeof(arp_data_t));
    return send_packet(&pkt);
}

/*
 * Send an ARP REMOVE request to the passive controller
 */
int arp_remove(arp_data_t *arp)
{
    syncpkt_t pkt;

    pkt.datatype = D_ARP;
    pkt.oper = OP_REMOVE;
    memcpy(&pkt.msg,(char *)arp,sizeof(arp_data_t));
    return send_packet(&pkt);
}

/*
 * Send an ARP FLUSH request to the passive controller
 */
int arp_flush(void)
{
    syncpkt_t pkt;

    pkt.datatype = D_ARP;
    pkt.oper = OP_FLUSH;
    memset(pkt.msg,0,sizeof(pkt.msg));
    return send_packet(&pkt);
}

/*
 * Update the passive controller ARP table from ours. Returns -1 on failure.
 */
int update_arp_table(void)
{
    arp_data_t arp;
    FILE *fp;
    char buf[256];
    char ip[16],mask[16];
    int b1,b2,b3,b4,errcount=0;

    if((fp = fopen("/proc/net/arp","r")) == 0) {
        sprintf(buf,"/proc/net/arp: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
        return(-1);
    }

    /* Flush the entire table and update all entries */
    arp_flush();

    if(fgets(buf,sizeof(buf),fp) != 0) {
        for(;fgets(buf,sizeof(buf),fp);) {
            if(sscanf(buf,"%s 0x%x 0x%x %s %s %s\n", ip, &arp.hwtype,
             &arp.flags, arp.hwaddr, mask, arp.dev) < 4)
                 break;
             arp.hwtype = htonl(arp.hwtype);
             arp.flags = htonl(arp.flags);
             /* parse out ip addr */
             sscanf(ip,"%d.%d.%d.%d",&b1,&b2,&b3,&b4);
             arp.ipaddr = htonl((b1<<24)+(b2<<16)+(b3<<8)+b4);
             if(mask[0] == '*')
                 arp.mask = 0;
             else {
                 sscanf(mask,"%d.%d.%d.%d",&b1,&b2,&b3,&b4);
                 arp.mask = htonl((b1<<24)+(b2<<16)+(b3<<8)+b4);
             }
             if(arp_add(&arp) != OP_ACK) errcount++;
        }
    }
    fclose(fp);
    return errcount;
}

/*
 * Send a ROUTE ADD request to the passive controller
 */
int route_add(route_data_t *rt)
{
    syncpkt_t pkt;

    pkt.datatype = D_ROUTE;
    pkt.oper = OP_ADD;
    memcpy(&pkt.msg,(char *)rt,sizeof(route_data_t));
    return send_packet(&pkt);
}

/*
 * Send a ROUTE REMOVE request to the passive controller
 */
int route_remove(route_data_t *rt)
{
    syncpkt_t pkt;

    pkt.datatype = D_ROUTE;
    pkt.oper = OP_REMOVE;
    memcpy(&pkt.msg,(char *)rt,sizeof(route_data_t));
    return send_packet(&pkt);
}

/*
 * Send a ROUTE FLUSH request to the passive controller
 */
int route_flush()
{
    syncpkt_t pkt;

    pkt.datatype = D_ROUTE;
    pkt.oper = OP_FLUSH;
    memset(pkt.msg,0,sizeof(pkt.msg));
    return send_packet(&pkt);
}

/*
 * Update the passive controller routing table from ours. 
 * Returns -1 on failure.
 */
int update_route_table(void)
{
    route_data_t rt;
    FILE *fp;
    char buf[256];
    int errcount=0;

    if((fp = fopen("/proc/net/route","r")) == 0) {
        sprintf(buf,"/proc/net/route: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
        return(-1);
    }

    /* Flush the entire table and update all entries */
    route_flush();

    if(fgets(buf,sizeof(buf),fp) != 0) {
        for(;fgets(buf,sizeof(buf),fp);) {
            if(sscanf(buf,"%s %x %x %x %d %d %d %x %d %d %d\n", rt.dev, 
             &rt.dest, &rt.gw, &rt.flags, &rt.refcnt, &rt.use, &rt.metric,
             &rt.mask, &rt.mtu, &rt.window, &rt.irtt) < 10)
                break;
            /* Dest, gw & mask are already in network byte order */
            rt.flags = htonl(rt.flags);
            rt.refcnt = htonl(rt.refcnt);
            rt.use = htonl(rt.use);
            rt.metric = htonl(rt.metric);
            rt.mtu = htonl(rt.mtu);
            rt.window = htonl(rt.window);
            rt.irtt = htonl(rt.irtt);
            if(route_add(&rt) != OP_ACK) errcount++;
        }
    }
    fclose(fp);
    return errcount;
}

/*
 * Send an IPFWD ADD request to the passive controller
 */
int ipfwd_add(ipfwd_data_t *ipfwd)
{
    syncpkt_t pkt;

    pkt.datatype = D_IPFWD;
    pkt.oper = OP_ADD;
    memcpy(&pkt.msg,(char *)ipfwd,sizeof(ipfwd_data_t));
    return send_packet(&pkt);
}

/*
 * Send an IPFWD REMOVE request to the passive controller
 */
int ipfwd_remove(ipfwd_data_t *ipfwd)
{
    syncpkt_t pkt;

    pkt.datatype = D_IPFWD;
    pkt.oper = OP_REMOVE;
    memcpy(&pkt.msg,(char *)ipfwd,sizeof(ipfwd_data_t));
    return send_packet(&pkt);
}

/*
 * Send an IPFWD FLUSH request to the passive controller
 */
int ipfwd_flush(void)
{
    syncpkt_t pkt;

    pkt.datatype = D_IPFWD;
    pkt.oper = OP_FLUSH;
    memset(&pkt.msg,0,sizeof(pkt.msg));
    return send_packet(&pkt);
}

/*
 * Update the passive controller ipfwd table from ours. 
 * Returns -1 on failure.
 */
int update_ipfwd_table(void)
{
    ipfwd_data_t ipf;
    FILE *fp;
    char buf[256];
    int i,errcount=0;

    if((fp = fopen("/proc/net/ip_forward","r")) == 0) {
        sprintf(buf,"/proc/net/ip_forward: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
        return(-1);
    }

    /* Flush the entire table and update all entries */
    ipfwd_flush();

    if(fgets(buf,sizeof(buf),fp) != 0) {
        for(;fgets(buf,sizeof(buf),fp);) {
            if(sscanf(buf,"%lX/%lX->%lX/%lX %s %lX %X %u %u %lu %lu %u %u %u %u %u %u %u %u %u %u A%X X%X\n",
             &ipf.saddr,&ipf.smask,&ipf.daddr,&ipf.dmask,ipf.vianame,
             &ipf.via_addr,&ipf.flags,&ipf.nsp,&ipf.ndp,&ipf.pcnt,&ipf.bcnt,
             &ipf.ports[0],&ipf.ports[1],&ipf.ports[2],&ipf.ports[3],
             &ipf.ports[4],&ipf.ports[5],&ipf.ports[6],&ipf.ports[7],
             &ipf.ports[8],&ipf.ports[9],&ipf.tosand,&ipf.tosxor) < 22)
                break;
            ipf.saddr = htonl(ipf.saddr);
            ipf.smask = htonl(ipf.smask);
            ipf.daddr = htonl(ipf.daddr);
            ipf.dmask = htonl(ipf.dmask);
            ipf.via_addr = htonl(ipf.via_addr);
            ipf.flags = htonl(ipf.flags);
            ipf.nsp = htonl(ipf.nsp);
            ipf.ndp = htonl(ipf.ndp);
            ipf.pcnt = htonl(ipf.pcnt);
            ipf.bcnt = htonl(ipf.bcnt);
            ipf.tosand = htonl(ipf.tosand);
            ipf.tosxor = htonl(ipf.tosxor);
            for(i=0;i<10;i++)
                ipf.ports[i] = htonl(ipf.ports[i]);
                
            if(ipfwd_add(&ipf) != OP_ACK) errcount++;
        }
    }
    fclose(fp);
    return errcount;
}

/*
 * Send an IPAUTOFW ADD request to the passive controller
 */
int ipautofw_add(ipautofw_data_t *ipf)
{
    syncpkt_t pkt;

    pkt.datatype = D_IPAUTOFW;
    pkt.oper = OP_ADD;
    memcpy(&pkt.msg,(char *)ipf,sizeof(ipautofw_data_t));
    return send_packet(&pkt);
}

/*
 * Send an IPAUTOFW REMOVE request to the passive controller
 */
int ipautofw_remove(ipautofw_data_t *ipf)
{
    syncpkt_t pkt;

    pkt.datatype = D_IPAUTOFW;
    pkt.oper = OP_REMOVE;
    memcpy(&pkt.msg,(char *)ipf,sizeof(ipautofw_data_t));
    return send_packet(&pkt);
}

/*
 * Send an IPAUTOFW FLUSH request to the passive controller
 */
int ipautofw_flush()
{
    syncpkt_t pkt;

    pkt.datatype = D_IPAUTOFW;
    pkt.oper = OP_FLUSH;
    memset(&pkt.msg,0,sizeof(ipautofw_data_t));
    return send_packet(&pkt);
}

/*
 * Update the passive controller ipautofw table from ours. 
 * Returns -1 on failure.
 */
int update_ipautofw_table(void)
{
    ipautofw_data_t ipf;
    FILE *fp;
    char buf[256];
    int errcount=0;

    if((fp = fopen("/proc/net/ip_autofw","r")) == 0) {
        sprintf(buf,"/proc/net/ip_autofw: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
        return(-1);
    }

    /* Flush the entire table and update all entries */
    ipautofw_flush();

    if(fgets(buf,sizeof(buf),fp) != 0) {
        for(;fgets(buf,sizeof(buf),fp);) {
            if(sscanf(buf,"%X %X %X-%X/%X %X %lX %lX %X %X %lu %X\n",
             &ipf.type,&ipf.proto,&ipf.low,&ipf.high,&ipf.visible,
             &ipf.hidden,&ipf.where,&ipf.lastcontact,&ipf.ctlproto,
             &ipf.ctlport,&ipf.timer_exp,&ipf.flags) < 11)
                break;
            ipf.type = htonl(ipf.type);
            ipf.proto = htonl(ipf.proto);
            ipf.low = htonl(ipf.low);
            ipf.high = htonl(ipf.high);
            ipf.visible = htonl(ipf.visible);
            ipf.hidden = htonl(ipf.hidden);
            ipf.where = htonl(ipf.where);
            ipf.lastcontact = htonl(ipf.lastcontact);
            ipf.ctlproto = htonl(ipf.ctlproto);
            ipf.ctlport = htonl(ipf.ctlport);
            ipf.timer_exp = htonl(ipf.timer_exp);
            ipf.flags = htonl(ipf.flags);
                
            if(ipautofw_add(&ipf) != OP_ACK) errcount++;
        }
    }
    fclose(fp);
    return errcount;
}
